// ZEVA Fuel Gauge Driver V2
// Open Source version, released under MIT License (see Readme file)
// Last modified by Ian Hooper (ZEVA), August 2021

// Code for ATtiny44A
// Fuses: Defaults except CKDIV8 off

// WARNING: Remember to select the current sensor size below

enum { HTFS200, HTFS400, HTFS800 };
#define CURRENT_SENSOR	HTFS200


#ifndef true
	#define true	1
	#define false	0
#endif

// Memory locations for settings
#define EEPROM_COUNTER_POS		4
#define	EEPROM_CAL_POS			8
#define	EEPROM_CAL10X_POS		12
#define EEPROM_WATERMARK_POS	16
#define EEPROM_SETTINGS_POS		20

// Used to detect when we're losing power, and should save state of charge to EEPROM
#define UNDERVOLTAGE_THRESHOLD	340	// 5V 0-1023 scale via 1:6 divider, translates to about 10V

//#define SHOW_SOC_ON_TACHO
#define BIDIRECTIONAL_WITH_4K_ZERO	false//true

// ADC channels
#define CURRENT_1X		1
#define CURRENT_10X		2
#define SUPPLY_VOLTAGE	4

// I/O pins
#define BUTTON1			(1<<PA6)
#define BUTTON2			(1<<PA5)
#define RED				(1<<PA7)
#define GREEN			(1<<PB0)
#define SENSOR_ENABLE	(1<<PA3)
#define TACH_OUT		(1<<PB1)

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/delay.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>

// Settings and running modes (i.e normal, or editing a certain setting)
enum { NORMAL = -1, PACK_CAPACITY, GAUGE_FULL, GAUGE_EMPTY, PULSES_PER_REV, NUM_SETTINGS };
unsigned char settings[NUM_SETTINGS] = { 0, 51, 0, 2 };
unsigned char minimums[NUM_SETTINGS] = { 0, 0, 0, 2 };
unsigned char maximums[NUM_SETTINGS] = { 50, 51, 51, 4 };

// Global variables
volatile short period;
volatile short ticks;
volatile short toggle = 0;
short zeroPoint = 0;
short zeroPoint10x = 0;
short deadband;
long counter = 0;
int prevSupplyVoltage = 0;
int adc[3][16];
int flashPackCapacity = -1;
int currentIsFlowing = false;

// Analog to Digital function
short ReadADC(unsigned char adc_input)
{
	ADMUX = adc_input; // No extra bits means use VCC as reference
	_delay_us(10); // Long delay to allow for changing of ADC system (single vs differential)
	ADCSRA |= 0x40; // Start the AD conversion
	while ((ADCSRA & 0x10) == 0); // Wait for the AD conversion to complete
	ADCSRA |= 0x10;
	return ADCW;
}

SIGNAL(TIM0_OVF_vect) // Interrupts at 15655Hz or so, used for main loop timing
{
	ticks++;
}

SIGNAL(TIM1_OVF_vect) // Generates variable frequency PWM for tach output
{
	ICR1 = period;
	if (toggle)
		PORTB |= TACH_OUT;
	else
		PORTB &= ~TACH_OUT;
	toggle = !toggle;
}

int Cap(int val, int min, int max)
{
	if (val > max)
		return max;
	else if (val < min)
		return min;
	return val;
}

int GetAvgADC(int n)
{
	int total = 0;
	for (int a=0; a<16; a++)
		total += adc[n][a];
	return total/16;
}

void SaveSoC()
{
	cli(); // Clear (disable) interrupts while writing to EEPROM
	eeprom_write_dword(EEPROM_COUNTER_POS, counter);
	eeprom_write_byte(0, 0); // Park EEPROM pointer new way - waits for all previous writing to finish
	sei(); // Enable interrupts again
}

void Calibrate()
{
	zeroPoint = 0;
	zeroPoint10x = 0;
	for (int n=0; n<16; n++)
	{
		_delay_ms(20);
		zeroPoint += ReadADC(CURRENT_1X); // was 1
		zeroPoint10x += ReadADC(CURRENT_10X); // was 2
	}
	zeroPoint /= 16;
	zeroPoint10x /= 16;

	cli();
	eeprom_write_word(EEPROM_CAL_POS, zeroPoint);
	eeprom_write_word(EEPROM_CAL10X_POS, zeroPoint10x);

	eeprom_write_byte(0, 0); // Park EEPROM pointer at sacrificial location
	sei();
}

void SaveSettings()
{
	cli();
	for (int n=0; n<NUM_SETTINGS; n++)
		eeprom_write_byte(EEPROM_SETTINGS_POS+n, settings[n]);

	eeprom_write_byte(0, 0); // Park EEPROM pointer
	sei();
}

int GetCurrent()
{
	// LEM 800A sensor, 800A = 1.25V
	// 1.5625 mV per amp
	// 4.883 mv per ADC step
	// VERY close to 3.125 amps per step
	// 3.125 is 25/8
	// But the 10x measurement is 0.3125 amps per step	

	short sensitiveCurrentRange = 1000; // tenths of an amp, i.e 100A
	short scaler = 1; // Default is for HTFS800
	if (CURRENT_SENSOR == HTFS200)
	{
		sensitiveCurrentRange = 250; // tenths of an amp
		scaler = 4; // compared with HTFS800, measurements would look 4x too big so have to scale down
	}
	else if (CURRENT_SENSOR == HTFS400)
	{
		sensitiveCurrentRange = 500; // tenths of an amp
		scaler = 2; // compared with HTFS800
	}
	
	short current = (GetAvgADC(CURRENT_1X) - zeroPoint)*31/scaler;
	if (current < sensitiveCurrentRange && current > -sensitiveCurrentRange)
		current = (zeroPoint10x - GetAvgADC(CURRENT_10X))*3/scaler; // inverted 10x signal

	if (current > -deadband && current < deadband) current = 0;

	return current;
}

int main(void)
{
	// Setup ports
	ADCSRA = 0b10000110; // top bit turns ADC on, the 110 LSBs gives a division factor of 64 (8mhz / 64 -> 125kHz ADC clock)
	DDRA = 0b10001000; // PA3: sensor enable, PA7: red LED
	DDRB = 0b00000111; // PB0: green LED, PB1: Tach out, PB2: Fuel gauge out
	PORTA |= BUTTON1 + BUTTON2; // Enable pull-ups on buttons

	PORTA |= SENSOR_ENABLE;

	// Timer/Counter 1 initialization
	TCCR1A = 0x02;
	TCCR1B = 0x18+0x04; // prescaler = /256, at 8Mhz so 31250Hz
	TIMSK1 |= (1<<TOIE1);
	period = 10000;
	ICR1 = period;
	//OCR1A = period>>1;

	// Timer/counter 2 initialisation
	TCCR0A = 0b10000001; // Phase correct PWM 0-0xFF-0, non-inverted PWM
	TCCR0B = 0b00000001; // Clock on, no prescaler
	TIMSK0 |= (1<<TOIE0); // Interrupt on overflow

	_delay_ms(100); // Brief delay to let inputs settle etc
	
	// Load data from EEPROM, if we have some saved (i.e if watermark is correct, to test EEPROM integrity)
	if (eeprom_read_dword(EEPROM_WATERMARK_POS) == 0xC0FFEE)
	{
		zeroPoint = eeprom_read_word(EEPROM_CAL_POS);
		zeroPoint10x = eeprom_read_word(EEPROM_CAL10X_POS);
		for (int n=0; n<NUM_SETTINGS; n++)
			settings[n] = Cap(eeprom_read_byte(EEPROM_SETTINGS_POS+n), minimums[n], maximums[n]); // Cap just to make sure

		counter = eeprom_read_dword(EEPROM_COUNTER_POS);
		//counter = (long)eeprom_read_byte(EEPROM_COUNTER_POS) * (long)settings[PACK_CAPACITY] * 36000L;

		//EEAR = 0; // Set EEPROM pointer to sacrificial location 0
		eeprom_write_byte(0, 0); // Park EEPROM pointer new way - waits for all previous writing to finish

	}
	else
	{
		eeprom_write_dword(EEPROM_WATERMARK_POS, 0xC0FFEE);
		Calibrate();
		SaveSettings();
	}

	deadband = 5; // Tenths of an amp
	if (CURRENT_SENSOR == HTFS400) deadband = 10;
	if (CURRENT_SENSOR == HTFS800) deadband = 20;

	sei(); // Enable interrupts
	wdt_enable(WDTO_500MS); // Enable watchdog timer

	int mode = NORMAL;
	int i = 0;
	int loop = 0;
	int button1timer = 0, button2timer = 0;
	int sleepTimer = 0;
	char didChangeMode = false;

	while (1)
	{
		wdt_reset();
		
		int supplyVoltage = ReadADC(SUPPLY_VOLTAGE);
		if (supplyVoltage < UNDERVOLTAGE_THRESHOLD && prevSupplyVoltage >= UNDERVOLTAGE_THRESHOLD)
		{
			SaveSoC(); // Quickly save SoC to EEPROM if we're losing power
		}
		prevSupplyVoltage = supplyVoltage;

		if (ticks > 98) // Happens at 160Hz
		{
			ticks -= 98;
			for (int n=1; n<=2; n++)
				adc[n][i] = ReadADC(n); // Digital oversampling of ADC for noise filtering

			if (!(PINA & BUTTON1))
				button1timer++;
			else
			{
				// check in here how long the button press was and act accordingly

				if (mode == NORMAL && button1timer > 160)
					counter = 0; // Reset SoC after 1 second press
				else if (button1timer > 8 && button2timer > 8) // Both buttons down
				{
					mode++;
					if (mode > PULSES_PER_REV)
					{
						mode = NORMAL;
						SaveSettings();
					}
					didChangeMode = true;
				}
				
				if (mode > NORMAL && button1timer > 8 && button2timer < 8 && !didChangeMode)
				{
					if (settings[mode] > minimums[mode]) settings[mode]--;
					if (mode == PACK_CAPACITY) flashPackCapacity = 0;
				}

				button1timer = 0;
			}

			if (!(PINA & BUTTON2))
				button2timer++;
			else
			{
				if (mode == NORMAL && button2timer > 160)
					Calibrate(); // Recalibrate current sensor after 1s press
				else if (button1timer > 8 && button2timer > 8) // Both buttons down
				{
					mode++;
					if (mode > PULSES_PER_REV)
					{
						mode = NORMAL;
						SaveSettings();
					}
					didChangeMode = true;
				}
				
				if (mode > NORMAL && button2timer > 8 && button1timer < 8 && !didChangeMode)
				{
					if (settings[mode] < maximums[mode]) settings[mode]++;
					if (mode == PACK_CAPACITY) flashPackCapacity = 0;
				}
				button2timer = 0;
			}

			if (button1timer == 0 && button2timer == 0) didChangeMode = false;

			i++;
		}


		if (i >= 16) // Main loop 10hz
		{
			i -= 16;

			loop++;
			if (loop >= 24) loop = 0;
			
			int current = GetCurrent();
			currentIsFlowing = !(current == 0);

			counter += current; // counter sees tenths of an amp added at 10hz, so units are hundredths of an amp-second

			if (counter >  200000000L) counter = 200000000L; // 200M 100ths of an amp-second = 2M amp-seconds = ~555Ah max
			if (counter < 0) counter = 0;

			int range = (settings[GAUGE_FULL] - settings[GAUGE_EMPTY])*5;

			long capacityInAmpSeconds = (long)settings[PACK_CAPACITY]*36000L;
			if (capacityInAmpSeconds == 0) capacityInAmpSeconds = 3600L; // Minimum is 1Ah, for testing
			long counterInAmpSeconds = counter / 100L;
			if (counterInAmpSeconds > capacityInAmpSeconds) counterInAmpSeconds = capacityInAmpSeconds;

			int position = 0;
			if (mode == GAUGE_FULL)
				position = settings[GAUGE_FULL]*5;
			else if (mode == GAUGE_EMPTY)
				position = settings[GAUGE_EMPTY]*5;
			else
				position = settings[GAUGE_FULL]*5 - (int)((long)range*counterInAmpSeconds/capacityInAmpSeconds);
				
			OCR0A = Cap(position, 1, 255);

			// Current of 300 needs to simulate 3000rpm so 50Hz, or 20ms per pulse at 1 pulse per rev, times scaler
			// PWM clock measured at 32000hz +/-1% so 640 PWM clocks for 300A @1p/rev
			// therefore clocks = 192000 / amps / pulsesPerRev
			// [Mar2013] - Changing to using timer to toggle pin, need to DOUBLE frequency, HALVE period, so use 192000/2 = 96000

			if (mode == PULSES_PER_REV) current = 1000; // Should display as 1000rpm, for tuning settings
			if (current < 0) current *= -1;
			if (current < 50) current = 50;
			if (current > 12000) current = 12000; // limit to 1200 amps i.e 12Krpm
			period = (960000L/(long)current) / settings[PULSES_PER_REV]; // added a zero because current is in tenths
		
			// Update LED
			PORTA &= ~RED;
			PORTB &= ~GREEN; // Start with both off
			if (mode == NORMAL)
			{
				// bottom bit of "loop" alterates at 10hz, second bit alternates at 5hz, third bit alternates at 2.5hz
				if (!currentIsFlowing || loop&0b100) // solid green if no current flowing, flash otherwise
					PORTB |= GREEN;
			}
			else
			{
				if (flashPackCapacity != -1) // then we're in the process of flashing pack capacity
				{
					if (flashPackCapacity > 5 && flashPackCapacity&2) // on 2, off 2, after a 0.5s delay
						PORTB |= GREEN;
					
					flashPackCapacity++;
					if (flashPackCapacity > settings[PACK_CAPACITY]*4+3)
						flashPackCapacity = -1; // Disable
				}
				else if (loop == 1 || (loop == 3 && mode > 0) || (loop == 5 && mode > 1) || (loop == 7 && mode > 2))
					PORTA |= RED;
			}			
		}
	}
}

